#vider la mmoire
rm(list=ls())

#fonction pour calcul du taux d'erreur
#et intervalle de confiance  90%
taux_erreur <- function(y_obs,y_pred){
  mc <- table(y_obs,y_pred)
  #affichage MC
  print("Matrice de confusion :")
  print(mc)
  #mal classs
  wrong <- sum(mc) - sum(diag(mc))
  print(paste("Mal classs :",wrong))
  #taux d'erreur
  err <- wrong/sum(mc)
  print(paste("Err :",err))
  #intervalle de confiance
  #ex. https://webapps.fundp.ac.be/umdb/biostats2017/biostat/modules/module105/page4.html
  #effectif
  n <- sum(mc)
  #cart-type
  et <- sqrt(err*(1-err)/n)
  #quantile loi normale 95%
  z <- qnorm(0.95)
  #borne basse
  bb <- err - z * et
  bh <- err + z * et
  #affichage bornes intervalle
  print("Bornes intervalle de confiance")
  print(c(bb,bh))
}

#chargement des donnes
DAll <- read.table(file="kr-vs-kp.txt",header=T,sep="\t")

#carac. de la base
str(DAll)

#*********************************
#partition apprentissage-test
#*********************************

#effectifs
nAll <- nrow(DAll)
nTrain <- 2196
nTest <- nAll-nTrain

#index pour partition
set.seed(1)
index <- sample(x=seq_len(nAll),size=nTrain)

#chantillon d'apprentissage
DTrain <- DAll[index,]
print(dim(DTrain))

#chantillon test
DTest <- DAll[-index,] #utilisation des indices ngatifs...
print(dim(DTest))

#*********************************
#modlisation - totalit des variables
#*********************************

#rgression logistique sur l'ensemble des variables
mAll <- glm(classe ~ ., data = DTrain, family = binomial)
print(mAll)

#performances
taux_erreur(DTest$classe,ifelse(predict(mAll,DTest,type="response")>0.5,"won","nowin"))

#*********************************
#slection backward
#*********************************

#slection backward BIC
library(MASS)
mBack <- stepAIC(mAll,direction="backward",k=log(nTrain),trace=FALSE)
print(mBack)

#nombre de variables slectionnes
print(length(mBack$coefficients)-1)

#liste des variables retenues dans le modle dfinitif
varBack <- sapply(tail(names(mBack$coefficients),length(mBack$coefficients)-1),function(x){substr(x,1,nchar(x)-1)})
names(varBack) <- NULL
print(varBack)

#ou plus simplement si on explore un peu les proprits de l'objet
#print(attr(mBack$terms,"term.labels"))

#ordre de retrait des variables
print(mBack$anova)

#volution du BIC en fonction du retrait des variables
#plot, rotation des labels
#https://stackoverflow.com/questions/5182238/replace-x-axis-with-own-values
plot(mBack$anova[,'AIC'],type="b",axes=FALSE,ylab="BIC",xlab="",main="Evolution BIC")
axis(1,at=1:length(mBack$anova[,'Step']),labels=mBack$anova[,'Step'],las=2,cex.axis=0.75)
axis(2)

#performances
#pas de diffrences significatives en performances du modle intgrant toutes les variables
#mais ce modle s'appuie avec moiti moins de variables
#plus facile  interprter, plus facile  dployer
taux_erreur(DTest$classe,ifelse(predict(mBack,DTest,type="response")>0.5,"won","nowin"))

#*********************************
#slection forward
#*********************************

#modele trivial
mNull <- glm(classe ~ 1, data = DTrain, family = binomial)

#slection forward BIC
mFwd <- stepAIC(mNull,scope=list(lower=formula(mNull),upper=formula(mAll)),direction="forward",k=log(nTrain),trace=FALSE)
print(mFwd)

#variables slectionnes
varFwd <- sapply(tail(names(mFwd$coefficients),length(mFwd$coefficients)-1),function(x){substr(x,1,nchar(x)-1)})
names(varFwd) <- NULL
print(varFwd)

#nombre de variables
print(length(varFwd))

#ordre d'introduction des variables
print(mFwd$anova)

#performances en test
#mmes perfs. que pour l'approche backward
taux_erreur(DTest$classe,ifelse(predict(mFwd,DTest,type="response")>0.5,"won","nowin"))

#*********************************
#tude des variables slectionnes
#*********************************

#variables en commun - forward et backward
print(intersect(varFwd,varBack))

#croisement des variables : prsent dans forward, mais pas dans backawrd
print(setdiff(varFwd,varBack))

#croisement des variables : prsent dans backward, mais pas dans forward
print(setdiff(varBack,varFwd))

#conclusion : la slection prsente un scnario de solution
#diffrents critres pour comparer : nombre de variables, performances en gnralisation
# galit sur les critres numriques : interprtabilit des rsultats

#*********************************
#slection par ranking
#*********************************

#package FSelector
library(FSelector)
importance <- FSelector::symmetrical.uncertainty(classe ~ ., data = DTrain)
print(importance)

#afficher dans l'ordre
idx <- order(importance$attr_importance,decreasing = TRUE)
sortedImp <- data.frame(var=rownames(importance)[idx],importance=importance$attr_importance[idx])
print(sortedImp)

#reprsentation graphique
barplot(rev(sortedImp$importance),horiz=TRUE,names.arg = rev(sortedImp$var),cex.names=0.6,las=2)

#rcuprons les max(foward,backward) meilleures variables 
selImportance <- FSelector::cutoff.k(importance,k=max(length(varFwd),length(varBack)))
print(selImportance)

#en commun avec le backward
print(intersect(selImportance,varBack))

#en commun avec le forward
print(intersect(selImportance,varFwd))

#convertir en formule
frmSelImp <- as.simple.formula(selImportance,"classe")
print(frmSelImp)

#lancer la rgression avec ces variables
mSelImp <- glm(frmSelImp, data = DTrain, family = binomial)
print(mSelImp)

#performances en test --  nombre de variables gales  backward et forward
#le modle est lgrement moins bon
#pas significativement (statistiquement, les intervalles de confiance se chevauchent)
#mais il y a quand mme 8 individus mal classs supplmentaires
#14 si on se rfre au modle initial comprenant toutes les variables
taux_erreur(DTest$classe,ifelse(predict(mSelImp,DTest,type="response")>0.5,"won","nowin"))

#*********************************
#slection CFS
#*********************************

#slection CFS
selCFS <- FSelector::cfs(classe ~ ., data = DTrain)

#nombre
print(length(selCFS))

#liste
print(selCFS)

#formule corresp.
frmCFS <- as.simple.formula(selCFS,"classe")
print(frmCFS)

#lancer la rgression avec ces variables
mCFS <- glm(frmCFS, data = DTrain, family = binomial)
print(mCFS)

#performances en test
#manifestement, l'approche CFS n'est absolument adapte  notre problme
#trop peu de variables slectionnes
taux_erreur(DTest$classe,ifelse(predict(mCFS,DTest,type="response")>0.5,"won","nowin"))

#*********************************
#ranking + wrapper
#*********************************

#p nombre de variables candidates
p <- ncol(DTrain)-1

#vecteur de taux de succes
succes <- matrix(0,nrow=p,ncol=2)

#utilisation de caret
library(caret)

#test des diffrentes configurations
for (j in 1:p){
  #j meilleures variables
  best <- cutoff.k(importance,k=j)
  #formule avec j variables
  formule <- as.simple.formula(best,"classe")
  print(formule)
  #modlisation avec caret
  lrm <- train(formule,data=DTrain,method="glm",trControl=trainControl(method="cv",number=5))
  #valuation
  succes[j,1] <- lrm$results$Accuracy
  succes[j,2] <- lrm$results$AccuracySD
}

#graphique des intervalles de confiance
#https://stackoverflow.com/questions/14069629/how-can-i-plot-data-with-confidence-intervals
plot(1:p,succes[,1],type="n",xlab="# de variables",ylab="Accuracy",main="Ranking + Wrapper (cross-validation)",ylim=c(0.65,1.0))
polygon(c(rev(1:p),1:p),c(rev(succes[,1]-succes[,2]),succes[,1]+succes[,2]),col = 'lavender', border = NA)
lines(1:p,succes[,1],type="l",lwd=2,col="lightgoldenrod4")
lines(1:p,succes[,1],type="b",lwd=2,col="lightgoldenrod4",cex=0.8)
lines(1:p,succes[,1]-succes[,2],type="l",col="lightpink3",lty=2)
lines(1:p,succes[,1]+succes[,2],type="l",col="lightpink3",lty=2)


#slection de la combinaison "optimale"
#id du max.
K.optim <- which.max(succes[,1])
print(K.optim)

#choisir les k meilleurs variables - mais toutes en fait dans notre exercice
bestK <- cutoff.k(importance,k=K.optim)
formuleK <- as.simple.formula(bestK,"classe")
print(formuleK)

#modliser
mBestK <- glm(formuleK, data = DTrain, family = binomial)

#valuer en test
taux_erreur(DTest$classe,ifelse(predict(mBestK,DTest,type="response")>0.5,"won","nowin"))
